# -*- coding: utf-8 -*-
"""Climate indices computation package based on Xarray."""
from xclim.core import units
from xclim.core.options import set_options
from xclim.indicators import ICCLIM, anuclim, atmos, icclim, land, seaIce

__author__ = """Travis Logan"""
__email__ = "logan.travis@ouranos.ca"
__version__ = "0.25.0"
